from typing import Optional
import tempfile


class RootDirectory:
    """
    This class is responsible for defining the monitoring the root directory for the cache.
    """
    def __init__(self, directory: Optional[str] = None) -> None:
        """
        The constructor for the RootDirectory.

        :param directory: (Optional[str]) the directory for the root of the cache
        """
        self._dir: str = self._extract_path() if directory is None else directory

    @staticmethod
    def _extract_path() -> str:
        """
        Extracts the first python path in the os environment variables.

        :return: (str)
        """
        return str(tempfile.TemporaryDirectory())

    @property
    def path(self) -> str:
        """
        Gets the path of the root of the directory.

        :return: (str) path to the root directory
        """
        return self._dir



"""this file defines the worker for managing local cache directories"""
import datetime
from uuid import UUID
import os
from typing import Optional
import shutil

from .errors import WorkerCacheError
from .register import Register


class Worker:
    """
    This is a class for managing a directory and meta data for temp files.

    Attributes:
        id (str): unique id for the worker
    """
    CLASS_BASE_DIR = os.path.dirname(os.path.realpath(__file__))

    def __init__(self, port: Optional[int], host: Optional[str],
                 existing_cache: Optional[str] = None, local_cache: Optional[str] = None) -> None:
        """
        The constructor for the Worker class.

        :param port: (Optional[int]) port for the redis connection tracking caches
        :param host: (Optional[str]) host for the redis connection tracking caches
        :param existing_cache: (Optional[str]) path to existing cache
        :param local_cache: (Optional[str]) path to the local cache
        """
        self._locked: bool = False
        self._port: Optional[int] = port
        self._host: Optional[str] = host
        # pylint: disable=invalid-name
        self.id: str = str(UUID(bytes=os.urandom(16), version=4))
        self._existing_cache: Optional[str] = existing_cache
        self.class_base_dir: str = self.CLASS_BASE_DIR if local_cache is None else local_cache
        self._base_dir: str = str(self.class_base_dir) + "/cache/{}/".format(self.id)
        self._connect_directory()
        if self._port is not None:
            Register(host=self._host, port=self._port).register_cache(cache_path=self.base_dir)

    @staticmethod
    def update_timestamp(cache_path: str) -> None:
        """
        Updates the cache timestamp.txt log with a new timestamp

        :param cache_path: (str) path to the cache being
        :return:
        """
        timestamp = datetime.datetime.now()
        file = open(cache_path + "timestamp.txt", "a")
        file.write("\n{}".format(timestamp))
        file.close()

    def lock(self) -> None:
        """
        Sets self._lock to True preventing cleanup.

        :return: None
        """
        self._locked = True

    def unlock(self) -> None:
        """
        Sets the self._lock to False enabling cleanup.

        :return: None
        """
        self._locked = False

    def _connect_directory(self) -> None:
        """
        Checks existing path, creates new cache if existing path not supplied.

        :return: None
        """
        if self._existing_cache is not None:
            if not os.path.isdir(self._existing_cache):
                raise WorkerCacheError(
                    message="directory '{}' was supplied as an existing cache but does not exist".
                    format(self._existing_cache)
                )
            self._base_dir = self._existing_cache
        else:
            self._generate_directory()

    def _generate_directory(self) -> None:
        """
        Generates cache directory with self.id (private).

        :return: None
        """
        if os.path.isdir(self._base_dir):
            raise WorkerCacheError(
                message="directory {} already exists. Check __del__ and self.id methods".
                format(self._base_dir)
            )
        os.makedirs(self._base_dir)
        self.update_timestamp(cache_path=self._base_dir)

    def _delete_directory(self) -> None:
        """
        Deletes cache directory (private).

        :return: None
        """
        if self._port is None and self._locked is False:
            shutil.rmtree(self.base_dir)
        elif self._port is not None:
            count: int = Register(host=self._host, port=self._port).deregister_cache(cache_path=self.base_dir,
                                                                                     locked=self._locked)
            if count == 0 and self._locked is False:
                shutil.rmtree(self.base_dir)

    @property
    def base_dir(self) -> str:
        """
        Dynamic property that defines the base directory for the cache.

        :return: (str) the base directory of the cache
        """
        return self._base_dir

    def __del__(self):
        """
        Fires when self is deleted, deletes the directory.

        :return: None
        """
        self._delete_directory()



"""this file defines the worker for pointing to caches in s3 buckets"""
import ast
from typing import Tuple, Optional, Any, Dict
from uuid import UUID
import os

import boto3
import botocore
import json


# pylint: disable=too-few-public-methods
class S3Worker:
    """
    This is a class for managing a directory for temp files in S3.

    Attributes:
        id (str): unique id for the worker
        base_dir (str): directory path for the cache
    """

    def __init__(self, cache_path: str, existing_cache: Optional[str] = None) -> None:
        """
        The constructor for the S3Worker class.

        :param cache_path: (str) the root path for all caches
        :param existing_cache: (Optional[str]) points to an existing cache if entered
        """
        # pylint: disable=invalid-name
        if existing_cache is None:
            self.id: str = str(UUID(bytes=os.urandom(16), version=4))
        else:
            self.id: str = self.extract_id(storage_path=existing_cache)
        self.base_dir: str = cache_path + "{}/".format(self.id)
        self._client: boto3.client = boto3.client('s3')
        self._resource: boto3.resource = boto3.resource('s3')
        self._locked: bool = False
        if existing_cache is None:
            self.create_meta()

    def delete_directory(self) -> None:
        """
        Deletes cache directory.

        :return: None
        """
        bucket, cache_path, short_file_name = self._split_s3_path(storage_path=self.base_dir)
        bucket = self._resource.Bucket(bucket)
        bucket.objects.filter(Prefix=cache_path).delete()

    def create_meta(self) -> None:
        """
        Creates the meta file for the cache.

        :return: None
        """
        bucket, cache_path, _ = self._split_s3_path(storage_path=self.base_dir)
        file_path = cache_path + 'meta.json'
        meta_object = self._resource.Object(bucket, file_path)
        meta_object.put(
            Body=(bytes(json.dumps({}).encode('UTF-8')))
        )

    def insert_meta(self, key: str, value: Any) -> None:
        """
        Inserts a value into the meta data file of the cache.

        :param key: (str) the key the value is denoted under
        :param value: (Any) the value to be inserted
        :return: None
        """
        bucket, cache_path, _ = self._split_s3_path(storage_path=self.base_dir)
        file_path = cache_path + 'meta.json'
        meta_object = self._resource.Object(bucket, file_path)
        meta_data = self.meta
        meta_data[key] = value

        meta_object.put(
            Body=(bytes(json.dumps(meta_data).encode('UTF-8')))
        )

    def lock(self) -> None:
        """
        Placeholder for locking as all S3 is locked.

        :return: None
        """
        pass

    def check_file(self, file: str) -> bool:
        """
        Checks to see if a file is present in the cache.

        :param file: (str) name of the file being checked
        :return: True if present, False if not
        """
        bucket, cache_path, _ = self._split_s3_path(storage_path=self.base_dir)
        try:
            file_path = cache_path + file
            self._resource.Object(bucket, file_path).load()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                return False
            else:
                raise
        return True

    @staticmethod
    def _split_s3_path(storage_path: str) -> Tuple[str, str, str]:
        """
        Splits the path to get parameters.

        :param storage_path: (str) path
        :return: bucket_name, file_name, short_file_name
        """
        path = storage_path.replace("s3://", "")
        path = path.split("/")
        bucket_name = path[0]
        file_name = "/".join(path[1:])
        short_file_name = path[-1]
        return bucket_name, file_name, short_file_name

    @staticmethod
    def extract_id(storage_path: str) -> str:
        """
        Extracts the id of the cache from a previous cache path.

        :param storage_path: (str) the cache path for the ID to be extracted
        :return: (str) the ID of the cache
        """
        return storage_path.split("/")[-1]

    @property
    def meta(self) -> Dict:
        """
        Extracts the meta data from the meta.json file of the cache.

        :return: (Dict) meta data of the cache
        """
        bucket, cache_path, _ = self._split_s3_path(storage_path=self.base_dir)
        file_path = cache_path + 'meta.json'
        meta_object = self._resource.Object(bucket, file_path)
        return ast.literal_eval(meta_object.get()['Body'].read().decode('utf-8'))



"""this file defines the errors"""
# pylint: disable=useless-super-delegation


class WorkerCacheError(Exception):
    """The error for the Worker class"""
    def __init__(self, message: str) -> None:
        """
        The constructor for the WorkerCacheError class.

        :param message: (str) the message for the error
        """
        super().__init__(message)


class CacheManagerError(Exception):
    """The error for the CacheManager class"""
    def __init__(self, message: str) -> None:
        """
        The constructor for the CacheManagerError class.

        :param message: (str) the message for the error
        """
        super().__init__(message)


class RegisterError(Exception):
    """The error for the Register class"""
    def __init__(self, message: str) -> None:
        """
        The constructor for the RegisterError class.

        :param message: (str) the message for the error
        """
        super().__init__(message)



"""This module manages the cache directories"""
import json
from typing import Union, Optional, Dict

from .errors import CacheManagerError
from .worker import Worker
from .s3_worker import S3Worker
from .root_directory import RootDirectory


class CacheManager:
    """
    This is a class for managing workers and meta data to a local cache.

    Attributes:
        worker (caching.worker.Worker): worker object that manages the access to the cache
        s3 (bool): if True, use S3 instead of local file storage
        s3_cache_path (Optional[str]): path to the cache in the s3
        local_cache_path (Optional[str]): path to the local cache
    """

    def __init__(self, port: Optional[int] = None, host: Optional[str] = None,
                 s3: bool = False, s3_cache_path: Optional[str] = None, local_cache_path: Optional[str] = None) -> None:
        """
        The constructor for the CacheManager class.

        :param port: (Optional[int]) port for the Redis connection to enable thread safe caching
        :param host: (Optional[str]) host for the Redis connection to enable thread safe caching
        :param s3: (bool) is True, connect to s3
        :param s3_cache_path: (Optional[str]) path to the cache in the s3
        :param local_cache_path: (Optional[str]) path to the local cache
        """
        self.worker: Union[None, Worker, S3Worker] = None
        # pylint: disable=invalid-name
        self.s3: bool = s3
        self._port: Optional[int] = port
        self._host: Optional[str] = host
        self.s3_cache_path: Optional[str] = s3_cache_path
        self.local_cache_path: Optional[str] = RootDirectory().path if local_cache_path is None else local_cache_path

    def create_cache(self, existing_cache: Optional[str] = None) -> None:
        """
        Deletes the old worker and creates a new one.

        :param existing_cache: (Optional[str]) path to existing cache
        :return: None
        """
        del self.worker
        if self.s3 is True:
            self.worker = S3Worker(cache_path=self.s3_cache_path, existing_cache=existing_cache)
        else:
            self.worker = Worker(port=self._port, host=self._host,
                                 existing_cache=existing_cache, local_cache=self.local_cache_path)
            if existing_cache is None:
                self._create_meta()
            else:
                if self.meta.get("locked", False) is True:
                    self.worker.lock()

    def lock_cache(self) -> None:
        """
        Locks the cache so it doesn't get wiped when finished.

        :return: None
        """
        if self.worker is None:
            raise CacheManagerError(message="cache worker is not defined so cannot be locked")
        if self.s3 is False:
            self.worker.lock()
            self.insert_meta(key="locked", value=True)

    def unlock_cache(self) -> None:
        """
        Unlocks the cache so it will get wiped when finished.

        :return: None
        """
        if self.worker is None:
            raise CacheManagerError(message="cache worker is not defined so cannot be unlocked")
        if self.s3 is False:
            self.worker.unlock()
            self.insert_meta(key="locked", value=False)

    def wipe_cache(self) -> None:
        """
        Deletes the current worker.

        :return: None
        """
        del self.worker
        self.worker = None

    def insert_meta(self, key: str, value: Union[str, int, float, dict, list]) -> None:
        """
        Inserts meta into the meta file in the cache.

        :param key: (str) key the value to be stored
        :param value: (Union[str, int, float, dict, list]) data to be stored
        :return: None
        """
        path = self.worker.base_dir + "meta.json"
        if self.s3 is False:
            with open(path) as meta_file:
                data = json.load(meta_file)

            data[key] = value

            with open(path, "w") as meta_file:
                json.dump(data, meta_file)
        else:
            self.worker.insert_meta(key=key, value=value)

    def _create_meta(self) -> None:
        """
        Writes an empty meta file.

        :return: None
        """
        if self.s3 is False:
            with open(self.worker.base_dir + "meta.json", "w") as meta_file:
                json.dump({}, meta_file)

    @property
    def cache_path(self):
        """
        Dynamic property.

        :return: if self.worker => self.worker.base_dir else None
        """
        if self.worker:
            return self.worker.base_dir
        return None

    @property
    def meta(self) -> Dict:
        """
        Dynamic property.

        :return: (dict) of meta data from cache
        """
        if self.s3 is False:
            with open(self.worker.base_dir + "meta.json") as meta_file:
                data = json.load(meta_file)
            return data
        return self.worker.meta

    def __del__(self):
        self.wipe_cache()

    def __enter__(self):
        self.create_cache()
        return self

    # pylint: disable=redefined-builtin
    def __exit__(self, type, value, traceback):
        self.wipe_cache()



import setuptools
from setuptools import find_packages
from setuptools.command.build_py import build_py as build_py_orig
import pathlib


class CustomBuildPy(build_py_orig):
    """
    subclass build_py so that we collect no .py files inside the built pip package
    this is done by overriding build_packages method with a noop
    """
    def build_packages(self):
        pass


with open("README.md", "r") as fh:
    long_description = fh.read()


with open('./requirements.txt', 'r') as requirements:
    requirements_buffer = requirements.read().split("\n")

with open(str(pathlib.Path(__file__).parent.absolute()) + "/monolithcaching/version.py", "r") as fh:
    version = fh.read().split("=")[1].replace("'", "")


directives = {
    'language_level': 3,
    'always_allow_keywords': True
}

setuptools.setup(
    name="monolithcaching",
    version=version,
    author="Maxwell Flitton",
    author_email="maxwell@monolithai.com",
    description="Python package for monolithcaching",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MonolithAILtd/caching",
    install_requires=requirements_buffer,
    packages=find_packages(exclude=("tests",)),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',
    tests_require=['pytest'],
    entry_points={
        'console_scripts': [
            'caching-hello = monolithcaching.console_commands.hello:print_logo',
        ],
    }
)



from typing import List, Optional

from redis import StrictRedis

from .errors import RegisterError


class Register:
    """
    This class is responsible for logging caches to Redis and counts how many caches are pointing to the cache.
    """
    TABLE = "CACHE_REGISTER"

    def __init__(self, host: str, port: int) -> None:
        """
        The constructor for the Register class.

        :param host: (str) host for the Redis connection
        :param port: (int) port for the Redis connection
        """
        self._connection: StrictRedis = StrictRedis(host=host, port=port, db=0)

    def get_count(self, cache_path: str) -> Optional[int]:
        """
        Gets the reference count for the cache.

        :param cache_path: (str) the path to the cache
        :return: (Optional[int]) number of count, or None if cache does not exist
        """
        count: Optional[str] = self._connection.hget(name=self.TABLE, key=cache_path)
        if count is None:
            return None
        return int(count)

    def register_cache(self, cache_path: str) -> None:
        """
        Registers the cache path if there isn't an entry or increasing the count by one if there is.

        :param cache_path: (str) the path to the cache
        :return: None
        """
        count: Optional[int] = self.get_count(cache_path=cache_path)
        if count is None:
            self._connection.hset(name=self.TABLE, key=cache_path, value="1")
        else:
            count += 1
            self._connection.hset(name=self.TABLE, key=cache_path, value=str(count))

    def deregister_cache(self, cache_path: str, locked: bool) -> int:
        """
        Deletes the cache path from Redis if the count goes to Zero or decreases the count by one if not zero.

        :param cache_path: (str) the path to the cache
        :param locked: (bool) is set to True prevents deleting of cache even if it has a count of zero
        :return: (int) the count of the references after the deregister.
        """
        count: Optional[int] = self.get_count(cache_path=cache_path)
        if count is None:
            raise RegisterError(message="cache {} is not in cache register so it cannot be deregistered".format(
                cache_path))
        else:
            if count > 0:
                count -= 1
            if count <= 0 and locked is False:
                self._connection.hdel(self.TABLE, cache_path)
            else:
                self._connection.hset(name=self.TABLE, key=cache_path, value=str(count))
        return count

    def get_all_records(self) -> List[dict]:
        """
        Get's all the key entries for the LOCAL_CLOUD.

        :return: (List[dict]) all cache entries
        """
        return self._connection.hgetall(name=self.TABLE)
